@@ -27,7 +27,7 @@
 /* #define DEBUG_STACK_SIZE */
 
 #ifndef VERSION
-#define VERSION "4.07"
+#define VERSION "4.11"
 #endif
 
 #ifndef USE_WIN32
@@ -37,14 +37,18 @@
 #include <floss.h>
 #endif
 
-/* POSIX threads */
-#if HAVE_PTHREAD_H && HAVE_LIBPTHREAD
-#define USE_PTHREAD
+/* threads model */
+#ifdef USE_UCONTEXT
+#define __MAKECONTEXT_V2_SOURCE
+#include <ucontext.h>
+#define DEBUG_STACK_SIZE
+#endif
+
+#ifdef USE_PTHREAD
 #define THREADS
 #define _REENTRANT
 #define _THREAD_SAFE
-#else
-#define USE_FORK
+#include <pthread.h>
 #endif
 
 /* TCP wrapper */
@@ -109,11 +113,13 @@ int _vsnprintf(char *, int, char *, ...);
 /* #define FD_SETSIZE 4096 */
 /* #define Win32_Winsock */
 #define __USE_W32_SOCKETS
-#include <windows.h>
 
 /* Winsock2 header for IPv6 definitions */
+#include <winsock2.h>
 #include <ws2tcpip.h>
 
+#include <windows.h>
+
 #define ECONNRESET WSAECONNRESET
 #define ENOTSOCK WSAENOTSOCK
 #define ENOPROTOOPT WSAENOPROTOOPT
@@ -255,6 +261,9 @@ extern char *sys_errlist[];
 #include <crypto.h> /* for CRYPTO_* and SSLeay_version */
 #endif
 
+/* CPU stack size */
+#define STACK_SIZE 65536
+
 /* I/O buffer size */
 #define BUFFSIZE        16384
 
