@@ -47,13 +47,7 @@ extern int num_clients;
 
 void main_initialize(char *, char *);
 void main_execute(void);
-void ioerror(char *);
-void sockerror(char *);
-void log_error(int, int, char *);
-char *my_strerror(int);
-int set_socket_options(int, int);
 char *stunnel_info(void);
-int alloc_fd(int);
 
 /**************************************** Prototypes for ssl.c */
 
@@ -82,24 +76,11 @@ void log_raw(const char *, ...)
 #else
     ;
 #endif
+void ioerror(const char *);
+void sockerror(const char *);
+void log_error(int, int, const char *);
+char *my_strerror(int);
     
-/**************************************** Prototypes for sthreads.c */
-
-typedef enum {
-    CRIT_KEYGEN, CRIT_INET, CRIT_CLIENTS, CRIT_WIN_LOG, CRIT_SESSION,
-    CRIT_SECTIONS
-} SECTION_CODE;
-
-void enter_critical_section(SECTION_CODE);
-void leave_critical_section(SECTION_CODE);
-void sthreads_init(void);
-unsigned long stunnel_process_id(void);
-unsigned long stunnel_thread_id(void);
-int create_client(int, int, void *, void *(*)(void *));
-#ifdef DEBUG_STACK_SIZE
-void stack_info(int);
-#endif
-
 /**************************************** Prototypes for pty.c */
 /* Based on Public Domain code by Tatu Ylonen <ylo@cs.hut.fi>  */
 
@@ -230,6 +211,33 @@ typedef struct {
 
 void parse_config(char *, char *);
 
+/**************************************** Prototypes for network.c */
+
+#define MAX_FD 64
+
+typedef struct {
+#ifdef HAVE_POLL
+    struct pollfd ufds[MAX_FD];
+    unsigned int nfds;
+#else
+    fd_set irfds, iwfds, orfds, owfds;
+    int max;
+#endif
+} s_poll_set;
+
+void s_poll_zero(s_poll_set *);
+void s_poll_add(s_poll_set *, int, int, int);
+int s_poll_canread(s_poll_set *, int);
+int s_poll_canwrite(s_poll_set *, int);
+int s_poll_wait(s_poll_set *, int);
+
+#ifndef USE_WIN32
+int signal_pipe_init(void);
+void exec_status(void);
+#endif
+int set_socket_options(int, int);
+int alloc_fd(int);
+
 /**************************************** Prototypes for client.c */
 
 typedef struct {
@@ -255,6 +263,7 @@ typedef struct {
     FD *sock_rfd, *sock_wfd; /* Read and write socket descriptors */
     FD *ssl_rfd, *ssl_wfd; /* Read and write SSL descriptors */
     int sock_bytes, ssl_bytes; /* Bytes written to socket and ssl */
+    s_poll_set fds; /* File descriptors */
 } CLI;
 
 extern int max_clients;
@@ -265,34 +274,8 @@ extern int max_fds;
 void *alloc_client_session(LOCAL_OPTIONS *, int, int);
 void *client(void *);
 
-/**************************************** Prototype for protocol.c */
-
-int negotiate(CLI *c);
-
 /**************************************** Prototypes for network.c */
 
-#define MAX_FD 64
-
-typedef struct {
-#ifdef HAVE_POLL
-    struct pollfd ufds[MAX_FD];
-    unsigned int nfds;
-#else
-    fd_set irfds, iwfds, orfds, owfds;
-    int max;
-#endif
-} s_poll_set;
-
-void s_poll_zero(s_poll_set *);
-void s_poll_add(s_poll_set *, int, int, int);
-int s_poll_canread(s_poll_set *, int);
-int s_poll_canwrite(s_poll_set *, int);
-int s_poll_wait(s_poll_set *, int);
-
-#ifndef USE_WIN32
-int signal_pipe_init(void);
-void exec_status(void);
-#endif
 int write_blocking(CLI *, int fd, u8 *, int);
 int read_blocking(CLI *, int fd, u8 *, int);
 /* descriptor versions of fprintf/fscanf */
@@ -308,10 +291,47 @@ int fdscanf(CLI *, int, const char *, char *)
 #else
        ;
 #endif
+
+/**************************************** Prototype for protocol.c */
+
+int negotiate(CLI *c);
+
+/**************************************** Prototypes for resolver.c */
+
 int name2addrlist(SOCKADDR_LIST *, char *, char *);
 int hostport2addrlist(SOCKADDR_LIST *, char *, char *);
 char *s_ntop(char *, SOCKADDR_UNION *);
 
+/**************************************** Prototypes for sthreads.c */
+
+typedef enum {
+    CRIT_KEYGEN, CRIT_INET, CRIT_CLIENTS, CRIT_WIN_LOG, CRIT_SESSION,
+    CRIT_SECTIONS
+} SECTION_CODE;
+
+void enter_critical_section(SECTION_CODE);
+void leave_critical_section(SECTION_CODE);
+void sthreads_init(void);
+unsigned long stunnel_process_id(void);
+unsigned long stunnel_thread_id(void);
+int create_client(int, int, void *, void *(*)(void *));
+#ifdef USE_UCONTEXT
+typedef struct CONTEXT_STRUCTURE {
+    char stack[STACK_SIZE];
+    unsigned long id;
+    ucontext_t ctx;
+    s_poll_set *fds;
+    int ready; /* number of ready file descriptors */
+    time_t finish; /* when to finish poll() for this context */
+    struct CONTEXT_STRUCTURE *next; /* next context on a list */
+} CONTEXT;
+extern CONTEXT *ready_head, *ready_tail;
+extern CONTEXT *waiting_head, *waiting_tail;
+#endif
+#ifdef DEBUG_STACK_SIZE
+void stack_info(int);
+#endif
+
 /**************************************** Prototypes for gui.c */
 
 #ifdef USE_WIN32
